import struct
from .error import SocketClosedError
from .bgp_message import BgpMessage

class Chopper(object):
    def __init__(self, input_stream):
        self.input_stream = input_stream

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        header, length, message_type = self.load_header()
        extra_data_length = length - BgpMessage.HEADER_LENGTH
        if extra_data_length > 0:
            serialised_body = self.input_stream.read(extra_data_length)
            if len(serialised_body) < extra_data_length:
                raise SocketClosedError(
                    "Tried to read %d bytes but only got %d" % (extra_data_length, len(header))
                )
        elif extra_data_length < 0:
            raise ValueError("Invalid BGP length field")
        else:
            serialised_body = b""

        return message_type, serialised_body

    def load_header(self):
        # TODO handle when stream runs out
        header = self.input_stream.read(19)
        if len(header) < 19:
            raise SocketClosedError("Tried to read %d bytes but only got %d" % (19, len(header)))

        marker, length, message_type = struct.unpack("!16sHB", header)

        if marker == BgpMessage.MARKER:
            return header, length, message_type

        raise ValueError("BGP marker missing")



import struct

def bytes_to_short(bytes_):
    """Convert bytes to short"""
    short, = struct.unpack("!H", bytes_)
    return short

def bytes_to_integer(bytes_):
    """Convert bytes to integer"""
    integer, = struct.unpack("!I", bytes_)
    return integer



import struct
import unittest

from beka.bgp_message import BgpMessage, BgpMessageParser, BgpMessagePacker, BgpOpenMessage
from beka.bgp_message import BgpUpdateMessage, BgpNotificationMessage, BgpKeepaliveMessage
from beka.ip import IP4Prefix, IP4Address
from beka.ip import IP6Prefix, IP6Address

def build_byte_string(hex_stream):
    values = [int(x, 16) for x in map(''.join, zip(*[iter(hex_stream)]*2))]
    return struct.pack("!" + "B" * len(values), *values)

class BgpMessageTestCase(unittest.TestCase):
    def test_open_message_parses_ipv4_multiprotocol(self):
        serialised_message = build_byte_string("04fe0900b4c0a8000f080206010400010001")
        message = BgpMessageParser().parse(BgpMessage.OPEN_MESSAGE, serialised_message)
        self.assertEqual(message.version, 4)
        self.assertEqual(message.peer_as, 65033)
        self.assertEqual(message.hold_time, 180)
        self.assertEqual(message.identifier, IP4Address.from_string("192.168.0.15"))
        self.assertEqual(message.capabilities["multiprotocol"], ["ipv4-unicast"])

    def test_open_message_parses_ipv6_multiprotocol(self):
        serialised_message = build_byte_string("04fe0900b4c0a8000f080206010400020001")
        message = BgpMessageParser().parse(BgpMessage.OPEN_MESSAGE, serialised_message)
        self.assertEqual(message.version, 4)
        self.assertEqual(message.peer_as, 65033)
        self.assertEqual(message.hold_time, 180)
        self.assertEqual(message.identifier, IP4Address.from_string("192.168.0.15"))
        self.assertEqual(message.capabilities["multiprotocol"], ["ipv6-unicast"])

    def test_open_message_parses_multiprotocol_4_byte_asn(self):
        serialised_message = build_byte_string("04fe0900b4c0a8000f0e020c01040001000141040000fdeb")
        message = BgpMessageParser().parse(BgpMessage.OPEN_MESSAGE, serialised_message)
        self.assertEqual(message.version, 4)
        self.assertEqual(message.peer_as, 65033)
        self.assertEqual(message.hold_time, 180)
        self.assertEqual(message.identifier, IP4Address.from_string("192.168.0.15"))
        self.assertEqual(message.capabilities["multiprotocol"], ["ipv4-unicast"])
        self.assertEqual(message.capabilities["fourbyteas"], [65003])

    def test_open_message_parses_route_refresh(self):
        serialised_message = build_byte_string("04fe0900b4c0a8000f0a02080104000200010200")
        message = BgpMessageParser().parse(BgpMessage.OPEN_MESSAGE, serialised_message)
        self.assertEqual(message.version, 4)
        self.assertEqual(message.peer_as, 65033)
        self.assertEqual(message.hold_time, 180)
        self.assertEqual(message.identifier, IP4Address.from_string("192.168.0.15"))
        self.assertEqual(message.capabilities["multiprotocol"], ["ipv6-unicast"])
        self.assertEqual(message.capabilities["routerefresh"], [True])

    def test_open_message_parses_exabgp_optional_params(self):
        serialised_message = build_byte_string("04fe0900b4c0a8000f9402060104000100010206010400010002020601040001000402060104000100800206010400010084020601040001008502060104000100860206010400020001020601040002000202060104000200040206010400020080020601040002008502060104000200860206010400190041020601040019004602060104400400470206010440040048020206000206410400010001")
        message = BgpMessageParser().parse(BgpMessage.OPEN_MESSAGE, serialised_message)
        self.assertEqual(message.version, 4)
        self.assertEqual(message.peer_as, 65033)
        self.assertEqual(message.hold_time, 180)
        self.assertEqual(message.identifier, IP4Address.from_string("192.168.0.15"))
        self.assertTrue("ipv4-unicast" in message.capabilities["multiprotocol"])
        self.assertTrue("ipv6-unicast" in message.capabilities["multiprotocol"])

    def test_open_message_packs(self):
        expected_serialised_message = build_byte_string("04fe0900b4c0a8000f080206010400010001")
        message = BgpOpenMessage(
            4, 65033, 180,
            IP4Address.from_string("192.168.0.15"),
            {"multiprotocol": ["ipv4-unicast"]}
        )
        serialised_message = BgpMessagePacker().pack(message)
        self.assertEqual(serialised_message[19:], expected_serialised_message)

    def test_open_message_packs_capabilities(self):
        expected_serialised_message = build_byte_string("04fe0900b4c0a8000f160214010400010001010400020001020041040000fdeb")
        capabilities = {
            "multiprotocol": ["ipv4-unicast", "ipv6-unicast"],
            "routerefresh": [True],
            "fourbyteas": [65003]
        }
        message = BgpOpenMessage(
            4, 65033, 180,
            IP4Address.from_string("192.168.0.15"),
            capabilities
        )
        serialised_message = BgpMessagePacker().pack(message)
        self.assertEqual(serialised_message[19:], expected_serialised_message)

    def test_keepalive_message_parses(self):
        serialised_message = b""
        message = BgpMessageParser().parse(BgpMessage.KEEPALIVE_MESSAGE, serialised_message)

    def test_keepalive_message_packs(self):
        expected_serialised_message = b""
        message = BgpKeepaliveMessage()
        serialised_message = BgpMessagePacker().pack(message)
        self.assertEqual(serialised_message[19:], expected_serialised_message)

    def test_notification_message_parses(self):
        serialised_message = build_byte_string("0202feb0")
        message = BgpMessageParser().parse(BgpMessage.NOTIFICATION_MESSAGE, serialised_message)
        self.assertEqual(message.error_code, 2)
        self.assertEqual(message.error_subcode, 2)
        self.assertEqual(message.data, b"\xfe\xb0")

    def test_notification_message_packs(self):
        expected_serialised_message = build_byte_string("0202feb0")
        message = BgpNotificationMessage(2, 2, b"\xfe\xb0")
        serialised_message = BgpMessagePacker().pack(message)
        self.assertEqual(serialised_message[19:], expected_serialised_message)

    def test_update_message_new_routes_parses(self):
        serialised_message = build_byte_string("0000000e40010101400200400304c0a80021080a")
        message = BgpMessageParser().parse(BgpMessage.UPDATE_MESSAGE, serialised_message)
        self.assertEqual(message.nlri[0], IP4Prefix.from_string("10.0.0.0/8"))
        self.assertEqual(
            message.path_attributes["next_hop"],
            IP4Address.from_string("192.168.0.33")
        )
        self.assertEqual(message.path_attributes["origin"], "EGP")
        self.assertEqual(message.path_attributes["as_path"], "")

    def test_update_message_new_routes_packs(self):
        expected_serialised_message = build_byte_string("0000000e40010101400200400304c0a80021080a17c0a840")
        nlri = [
            IP4Prefix.from_string("10.0.0.0/8"),
            IP4Prefix.from_string("192.168.64.0/23")
        ]
        path_attributes = {
            "next_hop": IP4Address.from_string("192.168.0.33"),
            "origin": "EGP",
            "as_path": ""
        }
        message = BgpUpdateMessage([], path_attributes, nlri)
        serialised_message = BgpMessagePacker().pack(message)
        self.assertEqual(serialised_message[19:], expected_serialised_message)

    def test_update_message_new_routes_parses_as4(self):
        serialised_message = build_byte_string("000000274001010040020802035ba0fe08fdebc0110e020300bc614e0000fe080000fdeb400304ac1900042009090909")
        message = BgpMessageParser().parse(BgpMessage.UPDATE_MESSAGE, serialised_message)
        self.assertEqual(message.nlri, [IP4Prefix.from_string("9.9.9.9/32")])
        self.assertEqual(message.path_attributes["next_hop"], IP4Address.from_string("172.25.0.4"))
        self.assertEqual(message.path_attributes["origin"], "IGP")
        self.assertEqual(message.path_attributes["as_path"], "23456 65032 65003")
        self.assertEqual(message.path_attributes["as4_path"], "12345678 65032 65003")

    def test_update_message_new_routes_packs_as4(self):
        expected_serialised_message = build_byte_string("000000274001010040020802035ba0fe08fdebc0110e020300bc614e0000fe080000fdeb400304ac1900042009090909")
        nlri = [
            IP4Prefix.from_string("9.9.9.9/32"),
        ]
        path_attributes = {
            "next_hop": IP4Address.from_string("172.25.0.4"),
            "origin": "IGP",
            "as_path": "23456 65032 65003",
            "as4_path": "12345678 65032 65003"
        }
        message = BgpUpdateMessage([], path_attributes, nlri)
        serialised_message = BgpMessagePacker().pack(message)
        self.assertEqual(serialised_message[19:], expected_serialised_message)

    def test_update_message_new_routes_parses_as4_new(self):
        serialised_message = build_byte_string("0000001c4001010040020e020300bc614e0000fe080001b2e5400304ac1900042009090909")
        parser = BgpMessageParser()
        parser.capabilities = {
            "fourbyteas": 12345
        }
        message = parser.parse(BgpMessage.UPDATE_MESSAGE, serialised_message)
        self.assertEqual(message.nlri, [IP4Prefix.from_string("9.9.9.9/32")])
        self.assertEqual(message.path_attributes["next_hop"], IP4Address.from_string("172.25.0.4"))
        self.assertEqual(message.path_attributes["origin"], "IGP")
        self.assertEqual(message.path_attributes["as_path"], "12345678 65032 111333")

    def test_update_message_new_routes_packs_as4_new(self):
        expected_serialised_message = build_byte_string("0000001c4001010040020e020300bc614e0000fe080001b2e5400304ac1900042009090909")
        nlri = [
            IP4Prefix.from_string("9.9.9.9/32"),
        ]
        path_attributes = {
            "next_hop": IP4Address.from_string("172.25.0.4"),
            "origin": "IGP",
            "as_path": "12345678 65032 111333"
        }
        message = BgpUpdateMessage([], path_attributes, nlri)
        packer = BgpMessagePacker()
        packer.capabilities = {"fourbyteas": 12345}
        serialised_message = packer.pack(message)
        self.assertEqual(serialised_message[19:], expected_serialised_message)

    def test_update_message_withdrawn_routes_parses(self):
        serialised_message = build_byte_string("0004180a01010000")
        message = BgpMessageParser().parse(BgpMessage.UPDATE_MESSAGE, serialised_message)
        self.assertEqual(message.withdrawn_routes[0], IP4Prefix.from_string("10.1.1.0/24"))

    def test_update_message_withdrawn_routes_packs(self):
        expected_serialised_message = build_byte_string("0004180a01010000")
        message = BgpUpdateMessage([IP4Prefix.from_string("10.1.1.0/24")], {}, [])
        serialised_message = BgpMessagePacker().pack(message)
        self.assertEqual(serialised_message[19:], expected_serialised_message)

    def test_update_v6_message_new_routes_parses(self):
        serialised_message = build_byte_string("0000004b400101004002040201fdeb800e3d0002012020010db80001000000000242ac110002fe800000000000000042acfffe110002007f20010db40000000000000000000000002f20010db30000")
        message = BgpMessageParser().parse(BgpMessage.UPDATE_MESSAGE, serialised_message)
        self.assertEqual(message.path_attributes["origin"], "IGP")
        self.assertEqual(
            message.path_attributes["mp_reach_nlri"]["next_hop"][0],
            IP6Address.from_string("2001:db8:1::242:ac11:2")
        )
        self.assertEqual(
            message.path_attributes["mp_reach_nlri"]["next_hop"][1],
            IP6Address.from_string("fe80::42:acff:fe11:2")
        )
        self.assertEqual(
            message.path_attributes["mp_reach_nlri"]["nlri"][0],
            IP6Prefix.from_string("2001:db4::/127")
        )
        self.assertEqual(
            message.path_attributes["mp_reach_nlri"]["nlri"][1],
            IP6Prefix.from_string("2001:db3::/47")
        )

    def test_update_v6_message_withdrawn_routes_parses(self):
        serialised_message = build_byte_string("0000002d800f2a0002017f20010db40000000000000000000000003020010db100003320010db20000002f20010db30000")
        message = BgpMessageParser().parse(BgpMessage.UPDATE_MESSAGE, serialised_message)

        expected_withdrawn_routes = [
            IP6Prefix.from_string("2001:db4::/127"),
            IP6Prefix.from_string("2001:db1::/48"),
            IP6Prefix.from_string("2001:db2::/51"),
            IP6Prefix.from_string("2001:db3::/47")
        ]

        self.assertEqual(message.path_attributes["mp_unreach_nlri"]["withdrawn_routes"], expected_withdrawn_routes)

    def test_update_v6_message_withdrawn_routes_packs(self):
        expected_serialised_message = build_byte_string("0000002d800f2a0002017f20010db40000000000000000000000003020010db100003320010db20000002f20010db30000")

        path_attributes = {
            "mp_unreach_nlri": {
                "withdrawn_routes": [
                    IP6Prefix.from_string("2001:db4::/127"),
                    IP6Prefix.from_string("2001:db1::/48"),
                    IP6Prefix.from_string("2001:db2::/51"),
                    IP6Prefix.from_string("2001:db3::/47")
                ]
            }
        }

        message = BgpUpdateMessage([], path_attributes, [])
        serialised_message = BgpMessagePacker().pack(message)
        self.assertEqual(serialised_message[19:], expected_serialised_message)

    def test_update_v6_message_new_routes_packs(self):
        expected_serialised_message = build_byte_string("0000004740010100400200800e3d0002012020010db80001000000000242ac110002fe800000000000000042acfffe110002007f20010db40000000000000000000000002f20010db30000")
        path_attributes = {
            "origin": "IGP",
            "as_path": "",
            "mp_reach_nlri": {
                "next_hop": [
                    IP6Address.from_string("2001:db8:1::242:ac11:2"),
                    IP6Address.from_string("fe80::42:acff:fe11:2")
                ],
                "nlri": [
                    IP6Prefix.from_string("2001:db4::/127"),
                    IP6Prefix.from_string("2001:db3::/47")
                ]
            }
        }
        message = BgpUpdateMessage([], path_attributes, [])
        serialised_message = BgpMessagePacker().pack(message)
        self.assertEqual(serialised_message[19:], expected_serialised_message)



import struct
import unittest
from io import BytesIO

from beka import chopper

class ChopperTestCase(unittest.TestCase):
    def test_valid_bgp_open_message(self):
        # test bgp open = "\x04\xfe\x09\x00\xb4\xc0\xa8\x00\x0f\x00"
        serialised_data = struct.pack(
            "!16sHB10s20s",
            b"\xFF" * 16,
            29,
            1,
            b"ten bytes!",
            b"junk data at the end"
        )
        expected_message = (1, b"ten bytes!")
        input_stream = BytesIO(serialised_data)
        message = chopper.Chopper(input_stream).next()

        self.assertEqual(message, expected_message)

    def test_bgp_message_invalid_marker(self):
        serialised_data = struct.pack(
            "!16sHB10s20s",
            b"\xFE" * 16,
            29,
            1,
            b"ten bytes!",
            b"junk data at the end"
        )
        input_stream = BytesIO(serialised_data)

        with self.assertRaises(ValueError) as context:
            chopper.Chopper(input_stream).next()

        self.assertEqual("BGP marker missing", str(context.exception))

    def test_bgp_message_invalid_length(self):
        serialised_data = struct.pack(
            "!16sHB10s20s",
            b"\xFF" * 16,
            17,
            1,
            b"ten bytes!",
            b"junk data at the end"
        )
        input_stream = BytesIO(serialised_data)

        with self.assertRaises(ValueError) as context:
            chopper.Chopper(input_stream).next()

        self.assertEqual("Invalid BGP length field", str(context.exception))



import time

from eventlet import sleep, GreenPool
from eventlet.queue import Queue
import eventlet.greenthread as greenthread

from .chopper import Chopper
from .event import EventTimerExpired, EventMessageReceived
from .bgp_message import BgpMessageParser, BgpMessagePacker
from .error import SocketClosedError, IdleError

class Peering(object):

    def __init__(self, state_machine, peer_address, socket, route_handler, error_handler=None):
        self.input_stream = None
        self.chopper = None
        self.pool = None
        self.eventlets = None
        self.parser = None
        self.packer = None
        self.state_machine = state_machine
        self.peer_address = peer_address[0]
        self.peer_port = peer_address[1]
        self.socket = socket
        self.route_handler = route_handler
        self.error_handler = error_handler
        self.start_time = int(time.time())

    def uptime(self):
        return int(time.time()) - self.start_time

    def run(self):
        self.input_stream = self.socket.makefile(mode="rb")
        self.chopper = Chopper(self.input_stream)
        self.pool = GreenPool()
        self.parser = BgpMessageParser()
        self.packer = BgpMessagePacker()
        self.state_machine.open_handler = self.open_handler
        self.eventlets = []

        self.eventlets.append(self.pool.spawn(self.send_messages))
        self.eventlets.append(self.pool.spawn(self.print_route_updates))
        self.eventlets.append(self.pool.spawn(self.kick_timers))
        self.eventlets.append(self.pool.spawn(self.receive_messages))

        self.pool.waitall()

    def open_handler(self, capabilities):
        self.parser.capabilities = capabilities
        self.packer.capabilities = capabilities

    def receive_messages(self):
        while True:
            sleep(0)
            try:
                message_type, serialised_message = self.chopper.next()
            except SocketClosedError as e:
                if self.error_handler:
                    self.error_handler("Peering %s: %s" % (self.peer_address, e))
                self.shutdown()
                break
            message = self.parser.parse(message_type, serialised_message)
            event = EventMessageReceived(message)
            tick = int(time.time())
            try:
                self.state_machine.event(event, tick)
            except IdleError as e:
                if self.error_handler:
                    self.error_handler("Peering %s: %s" % (self.peer_address, e))
                self.shutdown()
                break

    def send_messages(self):
        while True:
            sleep(0)
            message = self.state_machine.output_messages.get()
            self.socket.send(self.packer.pack(message))

    def empty_message_queue(self):
        while self.state_machine.output_messages.qsize():
            message = self.state_machine.output_messages.get()
            self.socket.send(self.packer.pack(message))

    def print_route_updates(self):
        while True:
            sleep(0)
            route_update = self.state_machine.route_updates.get()
            self.route_handler(route_update)

    def kick_timers(self):
        while True:
            sleep(1)
            tick = int(time.time())
            try:
                self.state_machine.event(EventTimerExpired(), tick)
            except IdleError as e:
                if self.error_handler:
                    self.error_handler("Peering %s: %s" % (self.peer_address, e))
                self.shutdown()
                break

    def shutdown(self):
        self.empty_message_queue()
        for eventlet in self.eventlets:
            eventlet.kill()



import struct
import unittest
from unittest.mock import MagicMock

from beka.bgp_message import BgpMessage, BgpOpenMessage, BgpUpdateMessage, BgpKeepaliveMessage, BgpNotificationMessage
from beka.state_machine import StateMachine
from beka.event import EventTimerExpired, EventMessageReceived, EventShutdown
from beka.ip import IP4Prefix, IP4Address
from beka.ip import IP6Prefix, IP6Address
from beka.route import RouteAddition, RouteRemoval
from beka.error import IdleError

def build_byte_string(hex_stream):
    values = [int(x, 16) for x in map(''.join, zip(*[iter(hex_stream)]*2))]
    return struct.pack("!" + "B" * len(values), *values)

def reset_timer_if_running(timer, tick):
    if timer.running():
        timer.reset(tick)

class StateMachinePassiveActiveTestCase(unittest.TestCase):
    def setUp(self):
        self.tick = 10000
        self.open_handler = MagicMock()
        self.state_machine = StateMachine(local_as=65001, peer_as=65002, local_address="1.1.1.1", router_id="1.1.1.1", neighbor="2.2.2.2", hold_time=240, open_handler=self.open_handler)
        self.assertEqual(self.state_machine.state, "active")
        self.assertEqual(self.state_machine.output_messages.qsize(), 0)
        self.assertFalse(self.state_machine.timers["hold"].running())
        self.assertFalse(self.state_machine.timers["keepalive"].running())

    def test_shutdown_message_advances_to_idle(self):
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventShutdown(), self.tick)
        self.assertEqual(self.state_machine.state, "idle")
        self.assertTrue("Shutdown requested" in str(context.exception))

    def test_timer_expired_event_does_nothing(self):
        self.tick += 3600
        self.state_machine.event(EventTimerExpired(), self.tick)
        self.assertEqual(self.state_machine.state, "active")
        self.assertFalse(self.state_machine.timers["hold"].expired(self.tick))
        self.assertFalse(self.state_machine.timers["keepalive"].expired(self.tick))
        self.assertEqual(self.state_machine.output_messages.qsize(), 0)
        self.assertEqual(self.state_machine.route_updates.qsize(), 0)

    def test_open_message_advances_to_open_confirm_and_sets_timers(self):
        capabilities = {"multiprotocol": "ipv4-unicast"}
        message = BgpOpenMessage(4, 65002, 240, IP4Address.from_string("2.2.2.2"), capabilities)
        self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "open_confirm")
        self.assertEqual(self.state_machine.output_messages.qsize(), 2)
        self.open_handler.assert_called_with(capabilities)
        self.assertTrue(isinstance(self.state_machine.output_messages.get(), BgpOpenMessage))
        self.assertTrue(isinstance(self.state_machine.output_messages.get(), BgpKeepaliveMessage))
        self.assertTrue(self.state_machine.timers["hold"].running())
        self.assertTrue(self.state_machine.timers["keepalive"].running())

    def test_keepalive_message_advances_to_idle(self):
        message = BgpKeepaliveMessage()
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "idle")

    def test_notification_message_advances_to_idle(self):
        message = BgpNotificationMessage(0, 0, b"")
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "idle")

    def test_update_message_advances_to_idle(self):
        path_attributes = {
            "next_hop" : IP4Address.from_string("5.4.3.2"),
            "as_path" : "65032 65011 65002",
            "origin" : "EGP"
            }
        message = BgpUpdateMessage([], path_attributes, [IP4Prefix.from_string("192.168.0.0/16")])
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "idle")

class StateMachineOpenSentTestCase(unittest.TestCase):
    def setUp(self):
        self.tick = 10000
        self.open_handler = MagicMock()
        self.state_machine = StateMachine(local_as=65001, peer_as=65002, local_address="1.1.1.1", router_id="1.1.1.1", neighbor="2.2.2.2", hold_time=240, open_handler=self.open_handler)
        self.assertEqual(self.state_machine.state, "active")
        # TODO trigger this correctly
        self.state_machine.state = "open_sent"
        self.state_machine.timers["hold"].reset(self.tick)
        self.assertEqual(self.state_machine.output_messages.qsize(), 0)
        self.assertFalse(self.state_machine.timers["keepalive"].running())
        self.assertTrue(self.state_machine.timers["hold"].running())

    def test_shutdown_message_advances_to_idle(self):
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventShutdown(), self.tick)
        self.assertEqual(self.state_machine.state, "idle")
        self.assertTrue("Shutdown requested" in str(context.exception))

    def test_hold_timer_expired_event_advances_to_idle_and_sends_notification(self):
        reset_timer_if_running(self.state_machine.timers["hold"], self.tick - 3600)
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventTimerExpired(), self.tick)
        self.assertEqual(self.state_machine.state, "idle")
        self.assertEqual(self.state_machine.output_messages.qsize(), 1)
        message = self.state_machine.output_messages.get()
        self.assertEqual(message.error_code, 4) # Hold Timer Expired

    def test_keepalive_timer_expired_event_does_nothing(self):
        reset_timer_if_running(self.state_machine.timers["keepalive"], self.tick - 3600)
        self.state_machine.event(EventTimerExpired(), self.tick)
        self.assertEqual(self.state_machine.state, "open_sent")
        self.assertEqual(self.state_machine.output_messages.qsize(), 0)
        self.assertEqual(self.state_machine.route_updates.qsize(), 0)

    def test_open_message_advances_to_open_confirm_and_sets_timers(self):
        capabilities = {"multiprotocol": "ipv4-unicast"}
        message = BgpOpenMessage(4, 65002, 240, IP4Address.from_string("2.2.2.2"), capabilities)
        self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "open_confirm")
        self.assertEqual(self.state_machine.output_messages.qsize(), 1)
        self.open_handler.assert_called_with(capabilities)
        self.assertTrue(isinstance(self.state_machine.output_messages.get(), BgpKeepaliveMessage))
        self.assertTrue(self.state_machine.timers["hold"].running())
        self.assertTrue(self.state_machine.timers["keepalive"].running())

    def test_keepalive_message_advances_to_idle(self):
        message = BgpKeepaliveMessage()
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "idle")

    def test_notification_message_advances_to_idle(self):
        message = BgpNotificationMessage(0, 0, b"")
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "idle")

    def test_update_message_advances_to_idle(self):
        path_attributes = {
            "next_hop" : IP4Address.from_string("5.4.3.2"),
            "as_path" : "65032 65011 65002",
            "origin" : "EGP"
            }
        message = BgpUpdateMessage([], path_attributes, [IP4Prefix.from_string("192.168.0.0/16")])
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "idle")

class StateMachineOpenConfirmTestCase(unittest.TestCase):
    def setUp(self):
        self.tick = 10000
        self.state_machine = StateMachine(local_as=65001, peer_as=65002, local_address="1.1.1.1", router_id="1.1.1.1", neighbor="2.2.2.2", hold_time=240)
        message = BgpOpenMessage(4, 65002, 240, IP4Address.from_string("2.2.2.2"), {"multiprotocol": "ipv4-unicast"})
        self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "open_confirm")
        for _ in range(self.state_machine.output_messages.qsize()):
            self.state_machine.output_messages.get()
        self.assertTrue(self.state_machine.timers["keepalive"].running())
        self.assertTrue(self.state_machine.timers["hold"].running())

    def test_shutdown_message_advances_to_idle_and_sends_notification(self):
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventShutdown(), self.tick)
        self.assertTrue("Shutdown requested" in str(context.exception))
        self.assertEqual(self.state_machine.state, "idle")
        self.assertEqual(self.state_machine.output_messages.qsize(), 1)
        message = self.state_machine.output_messages.get()
        self.assertEqual(message.error_code, 6) # Cease

    def test_hold_timer_expired_event_advances_to_idle_and_sends_notification(self):
        reset_timer_if_running(self.state_machine.timers["hold"], self.tick - 3600)
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventTimerExpired(), self.tick)
        self.assertEqual(self.state_machine.state, "idle")
        self.assertEqual(self.state_machine.output_messages.qsize(), 1)
        message = self.state_machine.output_messages.get()
        self.assertEqual(message.error_code, 4) # Hold Timer Expired

    def test_keepalive_timer_expired_event_sends_keepalive_and_resets_keepalive_timer(self):
        reset_timer_if_running(self.state_machine.timers["keepalive"], self.tick - 3600)
        self.state_machine.event(EventTimerExpired(), self.tick)
        self.assertEqual(self.state_machine.state, "open_confirm")
        self.assertEqual(self.state_machine.output_messages.qsize(), 1)
        message = self.state_machine.output_messages.get()
        self.assertTrue(isinstance(message, BgpKeepaliveMessage))
        self.assertFalse(self.state_machine.timers["keepalive"].expired(self.tick))
        self.assertTrue(self.state_machine.timers["keepalive"].running())

    def test_notification_message_advances_to_idle(self):
        message = BgpNotificationMessage(0, 0, b"")
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "idle")

    def test_keepalive_message_advances_to_established_and_resets_hold_timer(self):
        self.tick += 3600
        message = BgpKeepaliveMessage()
        self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "established")
        self.assertTrue(self.state_machine.timers["hold"].running())
        self.assertFalse(self.state_machine.timers["hold"].expired(self.tick))

    def test_keepalive_message_sends_all_routes(self):
        self.tick += 3600
        self.state_machine.routes_to_advertise = [
            RouteAddition(
                IP4Prefix.from_string("10.0.0.0/8"),
                IP4Address.from_string("192.168.1.33"),
                "",
                "IGP"
            ),
            RouteAddition(
                IP4Prefix.from_string("192.168.64.0/23"),
                IP4Address.from_string("192.168.1.33"),
                "",
                "IGP"
            ),
            RouteAddition(
                IP4Prefix.from_string("192.168.128.0/23"),
                IP4Address.from_string("192.168.1.34"),
                "",
                "IGP"
            )
        ]
        message = BgpKeepaliveMessage()
        self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "established")
        self.assertTrue(self.state_machine.timers["hold"].running())
        self.assertFalse(self.state_machine.timers["hold"].expired(self.tick))
        self.assertEqual(self.state_machine.output_messages.qsize(), 2)
        first_update = self.state_machine.output_messages.get()
        second_update = self.state_machine.output_messages.get()
        self.assertTrue(isinstance(first_update, BgpUpdateMessage))
        self.assertTrue(isinstance(second_update, BgpUpdateMessage))
        self.assertEqual(first_update.path_attributes["next_hop"], IP4Address.from_string("192.168.1.33"))
        self.assertEqual(first_update.nlri, [
            IP4Prefix.from_string("10.0.0.0/8"),
            IP4Prefix.from_string("192.168.64.0/23")
        ])
        self.assertEqual(second_update.path_attributes["next_hop"], IP4Address.from_string("192.168.1.34"))
        self.assertEqual(second_update.nlri, [
            IP4Prefix.from_string("192.168.128.0/23")
        ])

    def test_keepalive_message_sends_all_routes_v6(self):
        self.tick += 3600
        self.state_machine.routes_to_advertise = [
            RouteAddition(
                IP6Prefix.from_string("2001:db4::/127"),
                IP6Address.from_string("2001:db1::1"),
                "",
                "IGP"
            ),
            RouteAddition(
                IP6Prefix.from_string("2001:db5::/127"),
                IP6Address.from_string("2001:db1::1"),
                "",
                "IGP"
            ),
            RouteAddition(
                IP6Prefix.from_string("2001:db6::/127"),
                IP6Address.from_string("2001:db1::2"),
                "",
                "IGP"
            )
        ]
        message = BgpKeepaliveMessage()
        self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "established")
        self.assertTrue(self.state_machine.timers["hold"].running())
        self.assertFalse(self.state_machine.timers["hold"].expired(self.tick))
        self.assertEqual(self.state_machine.output_messages.qsize(), 2)
        first_update = self.state_machine.output_messages.get()
        second_update = self.state_machine.output_messages.get()
        self.assertTrue(isinstance(first_update, BgpUpdateMessage))
        self.assertTrue(isinstance(second_update, BgpUpdateMessage))
        self.assertEqual(first_update.path_attributes["mp_reach_nlri"]["next_hop"], [IP6Address.from_string("2001:db1::1")])
        self.assertEqual(first_update.path_attributes["mp_reach_nlri"]["nlri"], [
            IP6Prefix.from_string("2001:db4::/127"),
            IP6Prefix.from_string("2001:db5::/127")
        ])
        self.assertEqual(second_update.path_attributes["mp_reach_nlri"]["next_hop"], [IP6Address.from_string("2001:db1::2")])
        self.assertEqual(second_update.path_attributes["mp_reach_nlri"]["nlri"], [
            IP6Prefix.from_string("2001:db6::/127")
        ])

    def test_open_message_advances_to_idle_and_sends_notification(self):
        message = BgpOpenMessage(4, 65002, 240, IP4Address.from_string("2.2.2.2"), {"multiprotocol": "ipv4-unicast"})
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "idle")
        self.assertEqual(self.state_machine.output_messages.qsize(), 1)
        message = self.state_machine.output_messages.get()
        self.assertEqual(message.error_code, 6) # Cease

    def test_update_message_advances_to_idle(self):
        path_attributes = {
            "next_hop" : IP4Address.from_string("5.4.3.2"),
            "as_path" : "65032 65011 65002",
            "origin" : "EGP"
            }
        message = BgpUpdateMessage([], path_attributes, [IP4Prefix.from_string("192.168.0.0/16")])
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "idle")
        self.assertEqual(self.state_machine.output_messages.qsize(), 1)
        message = self.state_machine.output_messages.get()
        self.assertEqual(message.error_code, 5) # FSM error

class StateMachineEstablishedTestCase(unittest.TestCase):
    def setUp(self):
        self.tick = 10000
        self.state_machine = StateMachine(local_as=65001, peer_as=65002, local_address="1.1.1.1", router_id="1.1.1.1", neighbor="2.2.2.2", hold_time=240)
        message = BgpOpenMessage(4, 65002, 240, IP4Address.from_string("2.2.2.2"), {"multiprotocol": "ipv4-unicast"})
        self.state_machine.event(EventMessageReceived(message), self.tick)
        for _ in range(self.state_machine.output_messages.qsize()):
            self.state_machine.output_messages.get()
        message = BgpKeepaliveMessage()
        self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "established")
        self.assertTrue(self.state_machine.timers["hold"].running())
        self.assertTrue(self.state_machine.timers["keepalive"].running())

    def test_keepalive_timer_expired_event_sends_keepalive_and_resets_keepalive_timer(self):
        reset_timer_if_running(self.state_machine.timers["keepalive"], self.tick - 3600)
        self.state_machine.event(EventTimerExpired(), self.tick)
        self.assertEqual(self.state_machine.state, "established")
        self.assertEqual(self.state_machine.output_messages.qsize(), 1)
        message = self.state_machine.output_messages.get()
        self.assertTrue(isinstance(message, BgpKeepaliveMessage))
        self.assertTrue(self.state_machine.timers["keepalive"].running())
        self.assertFalse(self.state_machine.timers["keepalive"].expired(self.tick))

    def test_update_message_adds_route(self):
        path_attributes = {
            "next_hop" : IP4Address.from_string("5.4.3.2"),
            "as_path" : "65032 65011 65002",
            "origin" : "EGP"
        }
        route_attributes = {
            "prefix" : IP4Prefix.from_string("192.168.0.0/16"),
            "next_hop" : IP4Address.from_string("5.4.3.2"),
            "as_path" : "65032 65011 65002",
            "origin" : "EGP"
        }
        message = BgpUpdateMessage([], path_attributes, [IP4Prefix.from_string("192.168.0.0/16")])
        self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "established")
        self.assertEqual(self.state_machine.route_updates.qsize(), 1)
        self.assertEqual(self.state_machine.route_updates.get(), RouteAddition(**route_attributes))

    def test_update_message_removes_route(self):
        message = BgpUpdateMessage([IP4Prefix.from_string("192.168.0.0/16")], [], [])
        self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "established")
        self.assertEqual(self.state_machine.route_updates.qsize(), 1)
        self.assertEqual(self.state_machine.route_updates.get(), RouteRemoval(IP4Prefix.from_string("192.168.0.0/16")))

    def test_update_v6_message_adds_route(self):
        path_attributes = {
            "as_path" : "65032 65011 65002",
            "origin" : "EGP",
            "mp_reach_nlri" : {
                "next_hop" : [
                    IP6Address.from_string("2001:db8:1::242:ac11:2"),
                    IP6Address.from_string("fe80::42:acff:fe11:2"),
                ],
                "nlri" : [
                    IP6Prefix.from_string("2001:db4::/127"),
                ]
            }
        }
        route_attributes = {
            "prefix" : IP6Prefix.from_string("2001:db4::/127"),
            "next_hop" : IP6Address.from_string("2001:db8:1::242:ac11:2"),
            "as_path" : "65032 65011 65002",
            "origin" : "EGP"
        }
        message = BgpUpdateMessage([], path_attributes, [])
        self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "established")
        self.assertEqual(self.state_machine.route_updates.qsize(), 1)
        self.assertEqual(self.state_machine.route_updates.get(), RouteAddition(**route_attributes))

    def test_shutdown_message_advances_to_idle_and_sends_notification(self):
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventShutdown(), self.tick)
        self.assertTrue("Shutdown requested" in str(context.exception))
        self.assertEqual(self.state_machine.state, "idle")
        self.assertEqual(self.state_machine.output_messages.qsize(), 1)
        message = self.state_machine.output_messages.get()
        self.assertEqual(message.error_code, 6) # Cease

    def test_hold_timer_expired_event_advances_to_idle_and_sends_notification(self):
        reset_timer_if_running(self.state_machine.timers["hold"], self.tick - 3600)
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventTimerExpired(), self.tick)
        self.assertEqual(self.state_machine.state, "idle")
        self.assertEqual(self.state_machine.output_messages.qsize(), 1)
        message = self.state_machine.output_messages.get()
        self.assertEqual(message.error_code, 4) # Hold Timer Expired

    def test_notification_message_advances_to_idle(self):
        message = BgpNotificationMessage(0, 0, b"")
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "idle")

    def test_open_message_advances_to_idle_and_sends_notification(self):
        message = BgpOpenMessage(4, 65002, 240, IP4Address.from_string("2.2.2.2"), {"multiprotocol": "ipv4-unicast"})
        with self.assertRaises(IdleError) as context:
            self.state_machine.event(EventMessageReceived(message), self.tick)
        self.assertEqual(self.state_machine.state, "idle")
        self.assertEqual(self.state_machine.output_messages.qsize(), 1)
        message = self.state_machine.output_messages.get()
        self.assertEqual(message.error_code, 6) # Cease



class Timer:
    def __init__(self, count):
        self.count = count
        self.tick = None

    def running(self):
        """Return true if timer is set and running"""
        return bool(self.tick)

    def expired(self, tick):
        """Return true if time has elapsed"""
        return self.running() and tick > self.tick + self.count

    def reset(self, tick):
        """Reset time on timer"""
        self.tick = tick

    def stop(self):
        """Stop timer"""
        self.tick = None



from beka.ip import IPAddress, IPPrefix
from beka.ip import IP4Address, IP4Prefix
from beka.ip import IP6Address, IP6Prefix
import unittest

class IPAddressTestCase(unittest.TestCase):
    """Test IPAddress and its subclasses"""

    def make_addresses(self, address_strings):
        addresses = []

        for address_string in address_strings:
            addresses.append(IPAddress.from_string(address_string))

        return addresses

    def test_from_string(self):
        ipv4_address_strings = [
            "10.1.1.1",
            "10.1.0.0",
            "192.168.0.23"
        ]

        ipv6_address_strings = [
            "2404:138::1",
            "2404:138::",
            "2404:138::234:1",
            "::1"
        ]

        ipv4_addresses = self.make_addresses(ipv4_address_strings)
        ipv6_addresses = self.make_addresses(ipv6_address_strings)

        for address in ipv4_addresses:
            self.assertTrue(isinstance(address, IP4Address))
        for address in ipv6_addresses:
            self.assertTrue(isinstance(address, IP6Address))
        for address, address_string in zip(ipv4_addresses, ipv4_address_strings):
            self.assertEqual(str(address), address_string)
            self.assertEqual(repr(address), "IP4Address.from_string(\"%s\")" % address_string)
        for address, address_string in zip(ipv6_addresses, ipv6_address_strings):
            self.assertEqual(str(address), address_string)
            self.assertEqual(repr(address), "IP6Address.from_string(\"%s\")" % address_string)


class IPPrefixTestCase(unittest.TestCase):
    """Test IPPrefix and its subclasses"""

    def make_prefixes(self, prefix_strings):
        prefixes = []

        for prefix_string in prefix_strings:
            prefixes.append(IPPrefix.from_string(prefix_string))

        return prefixes

    def test_from_string(self):
        ipv4_prefix_strings = [
            "10.1.1.1/32",
            "10.1.0.0/16",
            "192.168.0.128/20"
        ]

        ipv6_prefix_strings = [
            "2404:138::1/128",
            "2404:138::/32",
            "2404:138::234:1/128",
            "::1/128"
        ]

        ipv4_prefixes = self.make_prefixes(ipv4_prefix_strings)
        ipv6_prefixes = self.make_prefixes(ipv6_prefix_strings)

        for prefix in ipv4_prefixes:
            self.assertTrue(isinstance(prefix, IP4Prefix))
        for prefix in ipv6_prefixes:
            self.assertTrue(isinstance(prefix, IP6Prefix))
        for prefix, prefix_string in zip(ipv4_prefixes, ipv4_prefix_strings):
            self.assertEqual(str(prefix), prefix_string)
            self.assertEqual(repr(prefix), "IP4Prefix.from_string(\"%s\")" % prefix_string)
        for prefix, prefix_string in zip(ipv6_prefixes, ipv6_prefix_strings):
            self.assertEqual(str(prefix), prefix_string)
            self.assertEqual(repr(prefix), "IP6Prefix.from_string(\"%s\")" % prefix_string)



from eventlet.queue import Queue
from collections import OrderedDict

from .event import Event
from .bgp_message import BgpMessage, BgpOpenMessage, BgpUpdateMessage
from .bgp_message import BgpKeepaliveMessage, BgpNotificationMessage
from .route import RouteAddition, RouteRemoval
from .ip import IPAddress, IPPrefix
from .ip import IP4Address, IP4Prefix
from .ip import IP6Address, IP6Prefix
from .timer import Timer
from .error import IdleError

class StateMachine:
    DEFAULT_HOLD_TIME = 240
    DEFAULT_KEEPALIVE_TIME = DEFAULT_HOLD_TIME // 3

    def __init__(self, local_as, peer_as, router_id, local_address, neighbor,
                 hold_time=DEFAULT_HOLD_TIME, open_handler=None):
        self.local_as = local_as
        if local_as > 65535:
            self.local_as2 = 23456
        else:
            self.local_as2 = self.local_as
        self.peer_as = peer_as
        self.router_id = IPAddress.from_string(router_id)
        self.local_address = IPAddress.from_string(local_address)
        self.neighbor = IPAddress.from_string(neighbor)
        self.hold_time = hold_time
        self.open_handler = open_handler

        self.keepalive_time = hold_time // 3
        self.output_messages = Queue()
        self.route_updates = Queue()
        self.routes_to_advertise = []
        self.fourbyteas = False

        self.timers = {
            "hold": Timer(self.hold_time),
            "keepalive": Timer(self.keepalive_time),
        }
        self.state = "active"

    def event(self, event, tick):
        if event.type == Event.TIMER_EXPIRED:
            self.handle_timers(tick)
        elif event.type == Event.MESSAGE_RECEIVED:
            self.handle_message(event.message, tick)
        elif event.type == Event.SHUTDOWN:
            self.handle_shutdown()

    def handle_shutdown(self):
        if self.state == "open_confirm" or self.state == "established":
            notification_message = BgpNotificationMessage(BgpNotificationMessage.CEASE)
            self.output_messages.put(notification_message)
        self.shutdown("Shutdown requested")

    def shutdown(self, message):
        self.state = "idle"
        raise IdleError("State machine stopping: %s" % message)

    def handle_timers(self, tick):
        if self.timers["hold"].expired(tick):
            self.handle_hold_timer()
        if self.timers["keepalive"].expired(tick):
            self.handle_keepalive_timer(tick)

    def handle_hold_timer(self):
        notification_message = BgpNotificationMessage(BgpNotificationMessage.HOLD_TIMER_EXPIRED)
        self.output_messages.put(notification_message)
        self.shutdown("Hold timer expired")

    def handle_keepalive_timer(self, tick):
        self.timers["keepalive"].reset(tick)
        message = BgpKeepaliveMessage()
        self.output_messages.put(message)

    def handle_message(self, message, tick):# state machine
        if self.state == "active":
            self.handle_message_active_state(message, tick)
        elif self.state == "open_sent":
            self.handle_message_open_sent_state(message, tick)
        elif self.state == "open_confirm":
            self.handle_message_open_confirm_state(message, tick)
        elif self.state == "established":
            self.handle_message_established_state(message, tick)

    def handle_message_active_state(self, message, tick):
        if isinstance(message, BgpOpenMessage):
            # TODO sanity check incoming open message
            if "fourbyteas" in message.capabilities:
                self.fourbyteas = message.capabilities["fourbyteas"]

            if self.open_handler:
                self.open_handler(message.capabilities)

            capabilities = {
                "fourbyteas": [self.local_as]
            }
            ipv4_capabilities = {"multiprotocol": ["ipv4-unicast"]}
            ipv6_capabilities = {"multiprotocol": ["ipv6-unicast"]}
            if isinstance(self.local_address, IP4Address):
                capabilities.update(ipv4_capabilities)
            elif isinstance(self.local_address, IP6Address):
                capabilities.update(ipv6_capabilities)
            open_message = BgpOpenMessage(4, self.local_as2, self.hold_time, self.router_id, capabilities)
            keepalive_message = BgpKeepaliveMessage()
            self.output_messages.put(open_message)
            self.output_messages.put(keepalive_message)
            self.timers["hold"].reset(tick)
            self.timers["keepalive"].reset(tick)
            self.state = "open_confirm"
        else:
            self.shutdown("Invalid message in Active state: %s" % str(message))

    def handle_message_open_sent_state(self, message, tick):
        if isinstance(message, BgpOpenMessage):
            # TODO sanity check incoming open message
            if "fourbyteas" in message.capabilities:
                self.fourbyteas = message.capabilities["fourbyteas"]

            if self.open_handler:
                self.open_handler(message.capabilities)

            capabilities = {
                "fourbyteas": [self.local_as]
            }
            ipv4_capabilities = {"multiprotocol": ["ipv4-unicast"]}
            ipv6_capabilities = {"multiprotocol": ["ipv6-unicast"]}
            if isinstance(self.local_address, IP4Address):
                capabilities.update(ipv4_capabilities)
            elif isinstance(self.local_address, IP6Address):
                capabilities.update(ipv6_capabilities)
            keepalive_message = BgpKeepaliveMessage()
            self.output_messages.put(keepalive_message)
            self.timers["hold"].reset(tick)
            self.timers["keepalive"].reset(tick)
            self.state = "open_confirm"
        else:
            self.shutdown("Invalid message in OpenSent state: %s" % str(message))

    def handle_message_open_confirm_state(self, message, tick):
        if isinstance(message, BgpKeepaliveMessage):
            for message in self.build_update_messages():
                self.output_messages.put(message)
            self.timers["hold"].reset(tick)
            self.timers["keepalive"].reset(tick)
            self.state = "established"
        elif isinstance(message, BgpNotificationMessage):
            self.shutdown("Notification message received %s" % str(message))
        elif isinstance(message, BgpOpenMessage):
            notification_message = BgpNotificationMessage(BgpNotificationMessage.CEASE)
            self.output_messages.put(notification_message)
            self.shutdown("Received Open message in OpenConfirm state")
        elif isinstance(message, BgpUpdateMessage):
            notification_message = BgpNotificationMessage(
                BgpNotificationMessage.FINITE_STATE_MACHINE_ERROR)
            self.output_messages.put(notification_message)
            self.shutdown("Received Update message in OpenConfirm state")

    def handle_message_established_state(self, message, tick):
        if isinstance(message, BgpUpdateMessage):
            self.process_route_update(message)
        elif isinstance(message, BgpKeepaliveMessage):
            self.timers["hold"].reset(tick)
        elif isinstance(message, BgpNotificationMessage):
            self.shutdown("Notification message received %s" % str(message))
        elif isinstance(message, BgpOpenMessage):
            notification_message = BgpNotificationMessage(BgpNotificationMessage.CEASE)
            self.output_messages.put(notification_message)
            self.shutdown("Received Open message in Established state")

    def process_route_update(self, update_message):
        # we handle both v4 and v6 here, in theory
        # this shouldn't happen in the real world though right?
        for prefix in update_message.nlri:
            route = RouteAddition(
                prefix,
                update_message.path_attributes["next_hop"],
                update_message.path_attributes["as_path"],
                update_message.path_attributes["origin"]
            )
            self.route_updates.put(route)
        if "mp_reach_nlri" in update_message.path_attributes:
            for prefix in update_message.path_attributes["mp_reach_nlri"]["nlri"]:
                route = RouteAddition(
                    prefix,
                    update_message.path_attributes["mp_reach_nlri"]["next_hop"][0],
                    update_message.path_attributes["as_path"],
                    update_message.path_attributes["origin"]
                )
                self.route_updates.put(route)
        for withdrawal in update_message.withdrawn_routes:
            route = RouteRemoval(
                withdrawal
            )
            self.route_updates.put(route)
        if "mp_unreach_nlri" in update_message.path_attributes:
            for withdrawal in update_message.path_attributes["mp_unreach_nlri"]["withdrawn_routes"]:
                route = RouteRemoval(
                    withdrawal
                )
                self.route_updates.put(route)

    def build_update_messages(self):
        # TODO handle withdrawals
        route_additions = list(filter(lambda x: isinstance(x, RouteAddition), self.routes_to_advertise))
        ipv4_route_additions = filter(lambda x: isinstance(x.prefix, IP4Prefix), route_additions)
        ipv6_route_additions = filter(lambda x: isinstance(x.prefix, IP6Prefix), route_additions)

        return self.build_ipv4_update_messages(ipv4_route_additions) + \
            self.build_ipv6_update_messages(ipv6_route_additions)

    def build_ipv4_update_messages(self, ipv4_route_additions):
        update_messages = []
        nlri_by_path = OrderedDict()
        for route_addition in ipv4_route_additions:
            path_key = (
                ("next_hop", route_addition.next_hop),
                ("as_path", route_addition.as_path),
                ("origin", route_addition.origin)
            )
            nlri_by_path.setdefault(path_key, []).append(route_addition.prefix)

        for path_attributes, nlri in nlri_by_path.items():
            update_messages.append(BgpUpdateMessage([], dict(path_attributes), nlri))

        return update_messages

    def build_ipv6_update_messages(self, ipv6_route_additions):
        update_messages = []
        nlri_by_path = OrderedDict()
        for route_addition in ipv6_route_additions:
            path_key = (
                ("next_hop", route_addition.next_hop),
                ("as_path", route_addition.as_path),
                ("origin", route_addition.origin)
            )
            nlri_by_path.setdefault(path_key, []).append(route_addition.prefix)

        for path_attributes, nlri in nlri_by_path.items():
            path_attributes_dict = dict(path_attributes)
            path_attributes_v6 = {
                "as_path": path_attributes_dict["as_path"],
                "origin": path_attributes_dict["origin"],
                "mp_reach_nlri": {
                    "next_hop": [
                        path_attributes_dict["next_hop"],
                    ],
                    "nlri": nlri
                }
            }
            update_messages.append(BgpUpdateMessage([], path_attributes_v6, []))

        return update_messages



import struct
import socket
from .packing_tools import bytes_to_short, bytes_to_integer
from .ip import IP4Prefix, IP4Address
from .ip import IP6Prefix, IP6Address
from io import BytesIO

class BgpMessage(object):
    OPEN_MESSAGE = 1
    UPDATE_MESSAGE = 2
    NOTIFICATION_MESSAGE = 3
    KEEPALIVE_MESSAGE = 4
    MARKER = b"\xFF" * 16
    HEADER_LENGTH = 19

PARSERS = {}

class BgpMessageParser(object):
    def __init__(self):
        self.capabilities = {}

    def parse(self, message_type, serialised_message):
        return PARSERS[message_type](serialised_message, self.capabilities)

def register_parser(cls):
    PARSERS[cls.MSG_TYPE] = cls.parse
    return cls

class BgpMessagePacker(object):
    def __init__(self):
        self.capabilities = {}

    def pack(self, message):
        packed_message = message.pack(self.capabilities)
        length = BgpMessage.HEADER_LENGTH + len(packed_message)
        header = struct.pack(
            "!16sHB",
            BgpMessage.MARKER,
            length,
            message.MSG_TYPE
            )
        return header + packed_message

MULTIPROTOCOL_TYPES = {
    (1, 1): "ipv4-unicast",
    (2, 1): "ipv6-unicast"
}

def parse_multiprotocol(serialised_capability):
    afi, reserved, safi = struct.unpack("!HBB", serialised_capability)
    if (afi, safi) in MULTIPROTOCOL_TYPES:
        return MULTIPROTOCOL_TYPES[(afi, safi)]
    return (afi, safi)

def parse_routerefresh(serialised_capability):
    return True

def parse_fourbyteas(serialised_capability):
    peer_as, = struct.unpack("!I", serialised_capability)
    return peer_as

capability_parsers = {
    1: parse_multiprotocol,
    2: parse_routerefresh,
    65: parse_fourbyteas
}

capability_keys = {
    1: "multiprotocol",
    2: "routerefresh",
    65: "fourbyteas",
}

def parse_capabilities(serialised_capabilities):
    stream = BytesIO(serialised_capabilities)
    capabilities = {}

    while True:
        serialised_header = stream.read(2)
        if len(serialised_header) == 0:
            break
        capability_code, capability_length = struct.unpack("!BB", serialised_header)
        serialised_capability = stream.read(capability_length)
        if capability_code in capability_keys:
            capability_key = capability_keys[capability_code]
            if capability_key not in capabilities:
                capabilities[capability_key] = []
            capabilities[capability_key].append(capability_parsers[capability_code](serialised_capability))
        else:
            print("WARNING did not recognise capability code %d" % capability_code)

    return capabilities

MULTIPROTOCOL_AFI_SAFI = {
    "ipv4-unicast": (1, 1),
    "ipv6-unicast": (2, 1),
}

def pack_multiprotocol(multiprotocol):
    afi, safi = MULTIPROTOCOL_AFI_SAFI[multiprotocol]
    return struct.pack("!HBB", afi, 0, safi)

def pack_routerefresh(route_refresh):
    return b""

def pack_fourbyteas(fourbyteas):
    return struct.pack("!I", fourbyteas)

capability_packers = {
    "multiprotocol": pack_multiprotocol,
    "routerefresh": pack_routerefresh,
    "fourbyteas": pack_fourbyteas,
}

capability_numbers = {
    "multiprotocol": 1,
    "routerefresh": 2,
    "fourbyteas": 65,
}

OPTIONAL_PARAMETER_CAPABILITY = 2

def pack_capabilities(capabilities):
    packed_capability_list = []

    for capability_key, capability_list in capabilities.items():
        capability_code = capability_numbers[capability_key]
        for capability in capability_list:
            packed_body = capability_packers[capability_key](capability)
            packed_header = struct.pack("!BB", capability_code, len(packed_body))
            packed_capability_list.append(packed_header + packed_body)

    packed_capabilities = b"".join(sorted(packed_capability_list))
    return packed_capabilities

def merge_dict_of_lists(main_dict, new_dict):
    for key, new_list in new_dict.items():
        if key not in main_dict:
            main_dict[key] = new_list
        else:
            main_dict[key] += new_list

def parse_optional_parameters(serialised_optional_parameters):
    stream = BytesIO(serialised_optional_parameters)
    capabilities = {}

    while True:
        serialised_header = stream.read(2)
        if len(serialised_header) == 0:
            break
        parameter_type, parameter_length = struct.unpack("!BB", serialised_header)
        if parameter_type != OPTIONAL_PARAMETER_CAPABILITY:
            raise ValueError("OPEN: Got unsupported optional parameter: %d" % parameter_type)
        serialised_capabilities = stream.read(parameter_length)

        next_capabilities = parse_capabilities(serialised_capabilities)
        merge_dict_of_lists(capabilities, next_capabilities)

    return capabilities

@register_parser
class BgpOpenMessage(BgpMessage):
    MSG_TYPE = BgpMessage.OPEN_MESSAGE

    def __init__(self, version, peer_as, hold_time, identifier, capabilities):
        self.version = version
        self.peer_as = peer_as
        self.hold_time = hold_time
        self.identifier = identifier
        self.capabilities = capabilities

    @classmethod
    def parse(cls, serialised_message, _capabilities):
        version, peer_as, hold_time, identifier, optional_parameters_length = struct.unpack(
            "!BHH4sB",
            serialised_message[:10]
        )
        capabilities = parse_optional_parameters(serialised_message[10:10+optional_parameters_length])
        return cls(version, peer_as, hold_time, IP4Address(identifier), capabilities)

    def pack(self, _capabilities):
        packed_capabilities = pack_capabilities(self.capabilities)
        capabilities_header = struct.pack(
            "!BB",
            OPTIONAL_PARAMETER_CAPABILITY,
            len(packed_capabilities))

        return struct.pack(
            "!BHH4sB",
            self.version,
            self.peer_as,
            self.hold_time,
            self.identifier.address,
            len(capabilities_header + packed_capabilities)
        ) + capabilities_header + packed_capabilities

    def __str__(self):
        return "BgpOpenMessage: Version %s, Peer AS: %s, Hold time: %s, Identifier: %s" % (
            self.version,
            self.peer_as,
            self.hold_time,
            self.identifier
            )

IP4_LENGTH = 4

def prefix_byte_length(bit_length):
    byte_length = bit_length // 8
    if bit_length % 8:
        byte_length += 1

    return byte_length

def pack_prefix(prefix, length):
    return prefix[:prefix_byte_length(length)]

def unpack_prefix(prefix):
    num_extra_bytes = IP4_LENGTH - len(prefix)

    if num_extra_bytes == 0:
        return prefix

    return prefix + b"\x00" * num_extra_bytes

def parse_nlri(serialised_nlri):
    stream = BytesIO(serialised_nlri)
    prefixes = []

    while True:
        serialised_length = stream.read(1)
        if len(serialised_length) == 0:
            break
        prefix_length = ord(serialised_length)
        packed_prefix = stream.read(prefix_byte_length(prefix_length))
        prefix = unpack_prefix(packed_prefix)
        prefixes.append(IP4Prefix(prefix, prefix_length))

    return prefixes

ORIGIN_CODES = {
    0: "IGP",
    1: "EGP",
    2: "INCOMPLETE"
}

def parse_origin(packed_origin):
    return ORIGIN_CODES[ord(packed_origin)]

AS_SET_CODE = 1
AS_SEQUENCE_CODE = 2
AS_NUMBER_LENGTH = 2
AS4_NUMBER_LENGTH = 4

def parse_as4_path(packed_as_path):
    # this does as_sets wrong, assumes everything is as_sequence
    input_stream = BytesIO(packed_as_path)
    as_numbers = []
    while True:
        packed_type_and_count = input_stream.read(2)
        if len(packed_type_and_count) == 0:
            break
        type_code, count = struct.unpack("!BB", packed_type_and_count)
        if type_code == AS_SET_CODE:
            print("WARNING received update with AS_SET, treating like AS_SEQUENCE")
        packed_as_sequence = input_stream.read(count * AS4_NUMBER_LENGTH)
        as_numbers += struct.unpack("!" + ("I" * count), packed_as_sequence)
    return " ".join(["%d" % x for x in as_numbers])

def parse_as_path(packed_as_path):
    # this does as_sets wrong, assumes everything is as_sequence
    input_stream = BytesIO(packed_as_path)
    as_numbers = []
    while True:
        packed_type_and_count = input_stream.read(2)
        if len(packed_type_and_count) == 0:
            break
        type_code, count = struct.unpack("!BB", packed_type_and_count)
        if type_code == AS_SET_CODE:
            print("WARNING received update with AS_SET, treating like AS_SEQUENCE")
        packed_as_sequence = input_stream.read(count * AS_NUMBER_LENGTH)
        as_numbers += struct.unpack("!" + ("H" * count), packed_as_sequence)
    return " ".join(["%d" % x for x in as_numbers])

def parse_next_hop(packed_next_hop):
    return IP4Address(packed_next_hop)

IP6_AFI = 2
UNICAST_SAFI = 1

IP6_LENGTH = 16

def unpack_prefix6(prefix):
    num_extra_bytes = IP6_LENGTH - len(prefix)

    if num_extra_bytes == 0:
        return prefix

    return prefix + b"\x00" * num_extra_bytes

def parse_nlri6(stream):
    prefixes = []

    while True:
        serialised_length = stream.read(1)
        if len(serialised_length) == 0:
            break
        prefix_length = ord(serialised_length)
        packed_prefix = stream.read(prefix_byte_length(prefix_length))
        prefix = unpack_prefix6(packed_prefix)
        prefixes.append(IP6Prefix(prefix, prefix_length))

    return prefixes

def parse_mp_reach_nlri(packed_mp_reach_nlri):
    attributes = {}
    stream = BytesIO(packed_mp_reach_nlri)
    afi, safi, next_hop_length = struct.unpack("!HBB", stream.read(4))
    if afi != IP6_AFI:
        raise ValueError("MP_REACH_NLRI: Got unsupported AFI: %d" % afi)
    if safi != UNICAST_SAFI:
        raise ValueError("MP_REACH_NLRI: Got unsupported SAFI: %d" % safi)
    if next_hop_length % IP6_LENGTH != 0:
        raise ValueError("MP_REACH_NLRI: Got unsupported next hop length: %d" % next_hop_length)

    attributes["next_hop"] = []
    for _ in range(next_hop_length // 16):
        attributes["next_hop"].append(IP6Address(struct.unpack("!16s", stream.read(IP6_LENGTH))[0]))

    _reserved_snpa = stream.read(1)

    attributes["nlri"] = parse_nlri6(stream)

    return attributes

def parse_mp_unreach_nlri(packed_mp_reach_nlri):
    attributes = {}
    stream = BytesIO(packed_mp_reach_nlri)
    afi, safi = struct.unpack("!HB", stream.read(3))
    if afi != IP6_AFI:
        raise ValueError("MP_UNREACH_NLRI: Got unsupported AFI: %d" % afi)
    if safi != UNICAST_SAFI:
        raise ValueError("MP_UNREACH_NLRI: Got unsupported SAFI: %d" % safi)

    attributes["withdrawn_routes"] = parse_nlri6(stream)

    return attributes

attribute_parsers = {
    1: parse_origin,
    2: parse_as_path,
    3: parse_next_hop,
    14: parse_mp_reach_nlri,
    15: parse_mp_unreach_nlri,
    17: parse_as4_path,
}

attribute_keys = {
    1: "origin",
    2: "as_path",
    3: "next_hop",
    14: "mp_reach_nlri",
    15: "mp_unreach_nlri",
    17: "as4_path",
}

ORIGIN_NUMBERS = {
    "IGP": 0,
    "EGP": 1,
    "INCOMPLETE": 2
}

def pack_origin(origin):
    return struct.pack("!B", ORIGIN_NUMBERS[origin])

def pack_as4_path(as_path):
    if not as_path:
        return b""

    as_numbers = [int(x) for x in as_path.split(" ")]
    count = len(as_numbers)
    header = struct.pack("!BB", AS_SEQUENCE_CODE, count)
    body = struct.pack("!" + ("I" * count), *as_numbers)
    return header + body

def pack_as_path(as_path):
    if not as_path:
        return b""

    as_numbers = [int(x) for x in as_path.split(" ")]
    count = len(as_numbers)
    header = struct.pack("!BB", AS_SEQUENCE_CODE, count)
    body = struct.pack("!" + ("H" * count), *as_numbers)
    return header + body

def pack_next_hop(next_hop):
    return next_hop.address

def pack_nlri6(nlri):
    packed_nlri = []

    for prefix in nlri:
        packed_prefix = struct.pack("!B", prefix.length) + pack_prefix(prefix.prefix, prefix.length)
        packed_nlri.append(packed_prefix)

    return b"".join(packed_nlri)

def pack_mp_reach_nlri(mp_reach_nlri):
    packed_header = struct.pack(
        "!HBB",
        IP6_AFI,
        UNICAST_SAFI,
        IP6_LENGTH * len(mp_reach_nlri["next_hop"])
    )
    packed_next_hop_list = []
    for next_hop in mp_reach_nlri["next_hop"]:
        packed_next_hop_list.append(struct.pack("!16s", next_hop.address))
    packed_next_hops = b"".join(packed_next_hop_list)
    packed_reserved = struct.pack("!B", 0)
    packed_nlri = pack_nlri6(mp_reach_nlri["nlri"])

    return packed_header + packed_next_hops + packed_reserved + packed_nlri

def pack_mp_unreach_nlri(mp_unreach_nlri):
    packed_header = struct.pack(
        "!HB",
        IP6_AFI,
        UNICAST_SAFI
    )
    packed_withdrawn_route_list = []
    for withdrawn_route in mp_unreach_nlri["withdrawn_routes"]:
        packed_prefix = struct.pack("!B", withdrawn_route.length) + \
            pack_prefix(withdrawn_route.prefix, withdrawn_route.length)
        packed_withdrawn_route_list.append(packed_prefix)
    packed_withdrawn_routes = b"".join(packed_withdrawn_route_list)

    return packed_header + packed_withdrawn_routes

attribute_packers = {
    "origin": pack_origin,
    "as_path": pack_as_path,
    "next_hop": pack_next_hop,
    "mp_reach_nlri" : pack_mp_reach_nlri,
    "mp_unreach_nlri" : pack_mp_unreach_nlri,
    "as4_path": pack_as4_path,
}

attribute_numbers = {
    "origin" : 1,
    "as_path" : 2,
    "next_hop" : 3,
    "mp_reach_nlri" : 14,
    "mp_unreach_nlri" : 15,
    "as4_path" : 17,
}

attribute_flags = {
    "origin" : 0x40,
    "as_path" : 0x40,
    "next_hop" : 0x40,
    "mp_reach_nlri" : 0x80,
    "mp_unreach_nlri" : 0x80,
    "as4_path" : 0xc0,
}

def parse_path_attributes(serialised_path_attributes, fourbyteas):
    stream = BytesIO(serialised_path_attributes)
    path_attributes = {}

    while True:
        attribute_header = stream.read(2)

        if len(attribute_header) == 0:
            break
        flags, type_code = struct.unpack("!BB", attribute_header)

        # TODO factor this pattern out
        if flags & 0x10:
            packed_length = stream.read(2)
            if len(packed_length) == 0:
                break
            length, = struct.unpack("!H", packed_length)
        else:
            packed_length = stream.read(1)
            if len(packed_length) == 0:
                break
            length, = struct.unpack("!B", packed_length)

        packed_attribute = stream.read(length)

        if type_code in attribute_parsers:
            # TODO we're ignoring the flags here, these should at very least be preserved
            # this is tightly coupled, there's gotta be a better way to do this
            if fourbyteas and attribute_keys[type_code] == "as_path":
                path_attributes["as_path"] = parse_as4_path(packed_attribute)
            else:
                path_attributes[attribute_keys[type_code]] = attribute_parsers[type_code](packed_attribute)
        else:
            print("WARNING did not recognise BGP path attribute type %d" % type_code)

    return path_attributes

def parse_withdrawn_routes(serialised_withdrawn_routes):
    stream = BytesIO(serialised_withdrawn_routes)
    prefixes = []

    while True:
        serialised_length = stream.read(1)
        if len(serialised_length) == 0:
            break
        prefix_length = ord(serialised_length)
        packed_prefix = stream.read(prefix_byte_length(prefix_length))
        prefix = unpack_prefix(packed_prefix)
        prefixes.append(IP4Prefix(prefix, prefix_length))

    return prefixes

PATH_ATTRIBUTE_ORDER = {
    "origin" : 1,
    "as_path" : 2,
    "as4_path" : 3,
    "next_hop" : 4,
    "mp_reach_nlri" : 5,
    "mp_unreach_nlri" : 6,
}

@register_parser
class BgpUpdateMessage(BgpMessage):
    MSG_TYPE = BgpMessage.UPDATE_MESSAGE

    def __init__(self, withdrawn_routes, path_attributes, nlri):
        self.withdrawn_routes = withdrawn_routes
        self.path_attributes = path_attributes
        self.nlri = nlri

    @classmethod
    def parse(cls, serialised_message, capabilities):
        data_stream = BytesIO(serialised_message)
        withdrawn_routes_length = bytes_to_short(data_stream.read(2))
        serialised_withdrawn_routes = data_stream.read(withdrawn_routes_length)
        withdrawn_routes = parse_withdrawn_routes(serialised_withdrawn_routes)

        total_path_attribute_length = bytes_to_short(data_stream.read(2))
        serialised_path_attributes = data_stream.read(total_path_attribute_length)
        path_attributes = parse_path_attributes(serialised_path_attributes, "fourbyteas" in capabilities)

        serialised_nlri = data_stream.read()
        nlri = parse_nlri(serialised_nlri)

        return cls(withdrawn_routes, path_attributes, nlri)

    def pack(self, capabilities):
        # TODO pack withdrawn routes
        packed_withdrawn_routes = self.pack_withdrawn_routes()
        packed_withdrawn_routes_length = struct.pack("!H", len(packed_withdrawn_routes))
        packed_path_attributes = self.pack_path_attributes("fourbyteas" in capabilities)
        packed_path_attributes_length = struct.pack("!H", len(packed_path_attributes))
        packed_nlri = self.pack_nlri()
        return packed_withdrawn_routes_length + \
            packed_withdrawn_routes + \
            packed_path_attributes_length + \
            packed_path_attributes + \
            packed_nlri

    def pack_withdrawn_routes(self):
        packed_routes_list = []

        for prefix in self.withdrawn_routes:
            packed_route = struct.pack("!B", prefix.length) + pack_prefix(prefix.prefix, prefix.length)
            packed_routes_list.append(packed_route)

        return b"".join(packed_routes_list)

    def pack_path_attributes(self, fourbyteas):
        packed_path_attributes = []
        sorted_attribute_pairs = sorted(self.path_attributes.items(), key=lambda x: PATH_ATTRIBUTE_ORDER[x[0]])
        for name, path_attribute in sorted_attribute_pairs:
            if fourbyteas and name=="as_path":
                packed_entry = pack_as4_path(path_attribute)
            else:
                packed_entry = attribute_packers[name](path_attribute)
            packed_header = struct.pack(
                "!BBB",
                attribute_flags[name],
                attribute_numbers[name],
                len(packed_entry)
            )
            packed_path_attribute = packed_header + packed_entry
            packed_path_attributes.append(packed_path_attribute)

        return b"".join(packed_path_attributes)

    def pack_nlri(self):
        packed_nlri = []

        for prefix in self.nlri:
            # TODO this feels like it should be on IP4Prefix
            packed_prefix = struct.pack("!B", prefix.length) + pack_prefix(prefix.prefix, prefix.length)
            packed_nlri.append(packed_prefix)

        return b"".join(packed_nlri)

    def __str__(self):
        return "BgpUpdateMessage: Withdrawn routes: %s, Path attributes: %s, NLRI: %s" % (
            [str(x) for x in self.withdrawn_routes],
            self.path_attributes,
            [str(x) for x in self.nlri]
            )

@register_parser
class BgpNotificationMessage(BgpMessage):
    MSG_TYPE = BgpMessage.NOTIFICATION_MESSAGE

    MESSAGE_HEADER_ERROR = 1
    OPEN_MESSAGE_ERROR = 2
    UPDATE_MESSAGE_ERROR = 3
    HOLD_TIMER_EXPIRED = 4
    FINITE_STATE_MACHINE_ERROR = 5
    CEASE = 6

    def __init__(self, error_code, error_subcode=0, data=b""):
        self.error_code = error_code
        self.error_subcode = error_subcode
        self.data = data

    @classmethod
    def parse(cls, serialised_message, _capabilities):
        error_code, error_subcode = struct.unpack("!BB", serialised_message[:2])
        data = serialised_message[2:]
        return cls(error_code, error_subcode, data)

    def pack(self, _capabilities):
        return struct.pack(
            "!BB",
            self.error_code,
            self.error_subcode,
        ) + self.data

    def __str__(self):
        return "BgpNotificationMessage: Error code: %s, Error subcode: %s, Data: %s" % (
            self.error_code,
            self.error_subcode,
            self.data
            )

@register_parser
class BgpKeepaliveMessage(BgpMessage):
    MSG_TYPE = BgpMessage.KEEPALIVE_MESSAGE

    def __init__(self):
        pass

    @classmethod
    def parse(cls, serialised_message, _capabilities):
        return cls()

    def pack(self, _capabilities):
        return b""

    def __eq__(self, other):
        return type(other) == type(self)

    def __str__(self):
        return "BgpKeepaliveMessage"



import signal
import sys
import yaml

from eventlet import GreenPool

from beka.beka import Beka


def printmsg(msg):
    sys.stderr.write("%s\n" % msg)
    sys.stderr.flush()


class Server():
    def __init__(self):
        self.peering_hosts = []
        self.greenlets = []
        self.bekas = []

    def run(self):
        signal.signal(signal.SIGINT, self.signal_handler)
        pool = GreenPool()

        with open("beka.yaml", encoding='utf-8') as file:
            config = yaml.safe_load(file.read())
        for router in config["routers"]:
            printmsg("Starting Beka on %s" % router["local_address"])
            beka = Beka(
                router["local_address"],
                router["bgp_port"],
                router["local_as"],
                router["router_id"],
                self.peer_up_handler,
                self.peer_down_handler,
                self.route_handler,
                self.error_handler
            )
            for peer in router["peers"]:
                beka.add_neighbor(
                    "passive",
                    peer["peer_ip"],
                    peer["peer_as"],
                )
            if "routes" in router:
                for route in router["routes"]:
                    beka.add_route(
                        route["prefix"],
                        route["next_hop"]
                    )
            self.bekas.append(beka)
            pool.spawn_n(beka.run)
        pool.waitall()
        printmsg("All greenlets gone, exiting")

    def signal_handler(self, _signal, _frame):
        printmsg("[SIGINT] Shutting down")
        self.shutdown()

    def shutdown(self):
        for beka in self.bekas:
            printmsg("Shutting down Beka %s" % beka)
            beka.shutdown()

    def peer_up_handler(self, peer_ip, peer_as):
        printmsg("[Peer up] %s %d" % (peer_ip, peer_as))

    def peer_down_handler(self, peer_ip, peer_as):
        printmsg("[Peer down] %s %s" % (peer_ip, peer_as))

    def error_handler(self, msg):
        printmsg("[Error] %s" % msg)

    def route_handler(self, route_update):
        if route_update.is_withdraw:
            printmsg("[Route handler] Route removed: %s" % route_update)
        else:
            printmsg("[Route handler] New route received: %s" % route_update)


if __name__ == "__main__":
    server = Server()
    server.run()



"""IP address and prefix classes"""

import socket

def is_ipv6(address_string):
    """Checks if a string is an IPv6 address or prefix"""

    return ":" in address_string

class IPBase(object): # pylint: disable=too-few-public-methods
    """Abstract base class for IP addresses and prefixes"""

    def __repr__(self):
        return "%s.from_string(\"%s\")" % (self.__class__.__name__, self.__str__())

class IPAddress(IPBase): # pylint: disable=too-few-public-methods
    """Abstract base class for IP addresses"""

    INET_TYPE = None

    def __init__(self, address):
        """Common constructor for IP addresses"""

        self.address = address

    def __str__(self):
        """Common str() for IP addresses"""

        address_string = socket.inet_ntop(self.INET_TYPE, self.address)
        return address_string

    def __eq__(self, other):
        """Common equality checker for IP addresses"""

        return self.address == other.address

    def __hash__(self):
        """Common hash method for IP addresses"""

        return hash(self.address)

    @staticmethod
    def from_string(string):
        """Common from_string method for IP addresses"""

        if is_ipv6(string):
            return IP6Address.build_from_string(string)
        return IP4Address.build_from_string(string)


class IPPrefix(IPBase): # pylint: disable=too-few-public-methods
    """Abstract base class for IP prefixes"""

    INET_TYPE = None

    def __init__(self, prefix, length):
        """Common constructor for IP prefixes"""

        self.prefix = prefix
        self.length = length

    def __str__(self):
        """Common str() for IP prefixes"""

        prefix_string = socket.inet_ntop(self.INET_TYPE, self.prefix)
        return "%s/%d" % (prefix_string, self.length)

    def __eq__(self, other):
        """Common equality checker for IP prefixes"""

        return self.prefix == other.prefix and self.length == other.length

    @staticmethod
    def from_string(string):
        """Common from_string method for IP prefixes"""

        if is_ipv6(string):
            return IP6Prefix.build_from_string(string)
        return IP4Prefix.build_from_string(string)


class IP4Address(IPAddress): # pylint: disable=too-few-public-methods
    """An IPv4 address"""

    INET_TYPE = socket.AF_INET

    @classmethod
    def build_from_string(cls, address_string):
        """Concrete build_from_string method"""

        address = socket.inet_pton(cls.INET_TYPE, address_string)

        return cls(address)

class IP4Prefix(IPPrefix): # pylint: disable=too-few-public-methods
    """An IPv4 prefix"""

    INET_TYPE = socket.AF_INET

    @classmethod
    def build_from_string(cls, string):
        """Concrete build_from_string method"""

        prefix_string, length_string = string.split("/")
        prefix = socket.inet_pton(cls.INET_TYPE, prefix_string)

        return cls(prefix, int(length_string, 10))

class IP6Address(IPAddress): # pylint: disable=too-few-public-methods
    """An IPv6 address"""

    INET_TYPE = socket.AF_INET6

    @classmethod
    def build_from_string(cls, address_string):
        """Concrete build_from_string method"""

        address = socket.inet_pton(socket.AF_INET6, address_string)

        return cls(address)

class IP6Prefix(IPPrefix): # pylint: disable=too-few-public-methods
    """An IPv6 prefix"""

    INET_TYPE = socket.AF_INET6

    @classmethod
    def build_from_string(cls, string):
        """Concrete build_from_string method"""

        prefix_string, length_string = string.split("/")
        prefix = socket.inet_pton(socket.AF_INET6, prefix_string)

        return cls(prefix, int(length_string, 10))



"""Exception classes for beka"""

class SocketClosedError(Exception):
    """Exception raised when socket closed error occurs"""

class IdleError(Exception):
    """Exception raised when idle error occurs"""



class RouteAddition:
    def __init__(self, prefix, next_hop, as_path, origin):
        self.prefix = prefix
        self.next_hop = next_hop
        self.as_path = as_path
        self.origin = origin
        self.is_withdraw = False

    def __str__(self):
        return "%s via %s (%s) %s" % (self.prefix, self.next_hop, self.as_path, self.origin)

    def __eq__(self, other):
        return self.prefix == other.prefix and \
            self.next_hop == other.next_hop and \
            self.as_path == other.as_path and \
            self.origin == other.origin

class RouteRemoval:
    def __init__(self, prefix):
        self.prefix = prefix
        self.next_hop = None
        self.is_withdraw = True

    def __str__(self):
        return str(self.prefix)

    def __eq__(self, other):
        return self.prefix == other.prefix



import unittest

from beka.beka import Beka
from beka.route import RouteAddition
from beka.ip import IPPrefix, IPAddress

class BekaTestCase(unittest.TestCase):
    def setUp(self):
        self.beka = Beka(
            local_address=None,
            bgp_port=None,
            local_as=None,
            router_id=None,
            peer_up_handler=None,
            peer_down_handler=None,
            route_handler=None,
            error_handler=None
        )

    def test_add_neighbor_must_be_passive(self):
        with self.assertRaises(ValueError) as context:
            self.beka.add_neighbor("active", "10.1.1.1", 65004)

        self.assertTrue("Only passive BGP supported" in str(context.exception))

    def test_add_neighbor_cannot_add_twice(self):
        self.assertFalse(self.beka.peers)
        self.beka.add_neighbor("passive", "10.1.1.1", 65004)
        self.assertTrue(self.beka.peers)
        self.assertEqual(self.beka.peers["10.1.1.1"], {
            "peer_ip": "10.1.1.1",
            "peer_as": 65004
        })

        with self.assertRaises(ValueError) as context:
            self.beka.add_neighbor("passive", "10.1.1.1", 65004)

        self.assertTrue("Peer already added" in str(context.exception))

    def test_add_route_adds_route(self):
        self.beka.add_route("10.1.0.0/16", "192.168.1.3")
        self.assertEqual(self.beka.routes_to_advertise[0],
            RouteAddition(
                prefix=IPPrefix.from_string("10.1.0.0/16"),
                next_hop=IPAddress.from_string("192.168.1.3"),
                as_path="",
                origin="IGP"
            )
        )

    def test_listening_on(self):
        beka = Beka(
            local_address='1.2.3.4',
            bgp_port=12345,
            local_as=23456,
            router_id='4.3.2.1',
            peer_up_handler=None,
            peer_down_handler=None,
            route_handler=None,
            error_handler=None
        )
        self.assertTrue(beka.listening_on('1.2.3.4', 12345))
        self.assertFalse(beka.listening_on('1.2.3.4', 23456))
        self.assertFalse(beka.listening_on('4.3.2.1', 12345))
        self.assertFalse(beka.listening_on('4.3.2.1', 23456))
        self.assertFalse(beka.listening_on('8.8.8.8', 53))




import unittest
from unittest.mock import patch, call

from eventlet import GreenPool, sleep
from eventlet.queue import Queue

from beka.peering import Peering


class RouteCatcher():  # pylint: disable=too-few-public-methods
    def __init__(self):
        self.route_updates = []

    def handle(self, route_update):
        self.route_updates.append(route_update)


class FakeStateMachine():  # pylint: disable=too-few-public-methods
    """Mocked StateMachine"""

    def __init__(self):
        self.output_messages = Queue()
        self.route_updates = Queue()


class FakeSocket():  # pylint: disable=too-few-public-methods
    """Mocked Socket"""

    def __init__(self):
        pass

    def makefile(self, *args, **kwargs):  # pylint: disable=unused-argument
        return None


class FakeChopper():  # pylint: disable=too-few-public-methods
    """Mocked Chopper"""

    def __init__(self):
        pass


class PeeringTestCase(unittest.TestCase):
    def setUp(self):
        self.route_catcher = RouteCatcher()
        self.state_machine = FakeStateMachine()
        self.peering = Peering(
            state_machine=self.state_machine,
            peer_address="1.2.3.4:179",
            socket=FakeSocket(),
            route_handler=self.route_catcher.handle
        )
        self.peering.chopper = FakeChopper()

    def test_print_route_updates(self):
        fake_route_update = "FAKE ROUTE UPDATE"
        self.state_machine.route_updates.put(fake_route_update)
        pool = GreenPool()
        eventlet = pool.spawn(self.peering.print_route_updates)
        for _ in range(10):
            sleep(0)
            if self.route_catcher.route_updates:
                break
        self.assertEqual(len(self.route_catcher.route_updates), 1)
        self.assertEqual(self.route_catcher.route_updates[0], fake_route_update)
        eventlet.kill()

    def test_run_starts_threads(self):
        with patch("beka.peering.GreenPool") as GreenPool:
            self.peering.run()
        GreenPool().spawn.assert_has_calls([
            call(self.peering.send_messages),
            call(self.peering.print_route_updates),
            call(self.peering.kick_timers),
            call(self.peering.receive_messages),
        ])
        assert GreenPool().waitall.call_count == 1



class Event():  # pylint: disable=too-few-public-methods
    TIMER_EXPIRED = 1
    MESSAGE_RECEIVED = 2
    SHUTDOWN = 3


class EventTimerExpired(Event):  # pylint: disable=too-few-public-methods
    """Event used to indicate timer has expired"""

    def __init__(self):
        # will work but please do this properly
        self.type = self.TIMER_EXPIRED


class EventMessageReceived(Event):  # pylint: disable=too-few-public-methods
    """Event used to indicate message has been received"""

    def __init__(self, message):
        # will work but please do this properly
        self.type = self.MESSAGE_RECEIVED
        self.message = message


class EventShutdown(Event):  # pylint: disable=too-few-public-methods
    """Event used to indicate shutdown"""

    def __init__(self):
        # will work but please do this properly
        self.type = self.SHUTDOWN



from copy import copy

from .stream_server import StreamServer

from .state_machine import StateMachine
from .peering import Peering
from .route import RouteAddition, RouteRemoval
from .ip import IPAddress, IPPrefix

DEFAULT_BGP_PORT = 179

class Beka(object):
    def __init__(self, local_address, bgp_port, local_as,
            router_id, peer_up_handler, peer_down_handler,
            route_handler, error_handler):
        self.local_address = local_address
        self.bgp_port = bgp_port
        self.local_as = local_as
        self.router_id = router_id
        self.peer_up_handler = peer_up_handler
        self.peer_down_handler = peer_down_handler
        self.route_handler = route_handler
        self.error_handler = error_handler

        self.peers = {}
        self.peerings = []
        self.stream_server = None
        self.routes_to_advertise = []

        if not self.bgp_port:
            self.bgp_port = DEFAULT_BGP_PORT

    def add_neighbor(self, connect_mode, peer_ip, peer_as):
        if connect_mode != "passive":
            raise ValueError("Only passive BGP supported")
        if peer_ip in self.peers:
            raise ValueError("Peer already added: %s %d" % (peer_ip, peer_as))

        self.peers[peer_ip] = {
            "peer_ip": peer_ip,
            "peer_as": peer_as
        }

    def add_route(self, prefix, next_hop):
        self.routes_to_advertise.append(
            RouteAddition(
                prefix=IPPrefix.from_string(prefix),
                next_hop=IPAddress.from_string(next_hop),
                as_path="",
                origin="IGP"
            )
        )

    def neighbor_states(self):
        states = []
        for peering in self.peerings:
            states.append((
                peering.peer_address,
                {
                    'info': {
                        'uptime': peering.uptime()
                    }
                }
            ))

        return states

    def run(self):
        self.stream_server = StreamServer((self.local_address, self.bgp_port), self.handle)
        self.stream_server.serve_forever()

    def handle(self, socket, address):
        peer_ip = address[0]
        if peer_ip not in self.peers:
            if self.error_handler:
                self.error_handler("Rejecting connection from %s:%d" % address)
            socket.close()
            return
        peer = self.peers[peer_ip]
        state_machine = StateMachine(
            local_as=self.local_as,
            peer_as=peer["peer_as"],
            router_id=self.router_id,
            local_address=self.local_address,
            neighbor=peer["peer_ip"]
        )
        state_machine.routes_to_advertise = copy(self.routes_to_advertise)
        peering = Peering(state_machine, address, socket, self.route_handler, error_handler=self.error_handler)
        self.peerings.append(peering)
        self.peer_up_handler(peer_ip, peer["peer_as"])
        peering.run()
        self.peer_down_handler(peer_ip, peer["peer_as"])
        self.peerings.remove(peering)

    def shutdown(self):
        if self.stream_server:
            self.stream_server.stop()
        for peering in self.peerings:
            peering.shutdown()

    def listening_on(self, address, port):
        return self.local_address == address and self.bgp_port == port



import socket

from eventlet import GreenPool, listen
import eventlet.greenthread as greenthread


class StreamServer:
    def __init__(self, address, handler):
        self.address = address
        self.handler = handler
        self.greenlets = set()
        self.running = False
        self.server = None

    def _family(self):
        if ":" in self.address[0]:
            return socket.AF_INET6
        return socket.AF_INET

    def serve_forever(self):
        self.running = True
        self.server = listen(self.address, self._family())
        pool = GreenPool()

        try:
            while self.running:
                sock, address = self.server.accept()
                pool.spawn(self.call_handler, sock, address)
                self.greenlets
        except OSError:
            pass

    def call_handler(self, sock, address):
        self.greenlets.add(greenthread.getcurrent())
        self.handler(sock, address)
        self.greenlets.remove(greenthread.getcurrent())

    def stop(self):
        self.running = False
        for greenlet in self.greenlets:
            greenlet.kill()
        self.server.shutdown(socket.SHUT_RDWR)
